#include <avr/sleep.h>
#define  F_CPU  4000000 // 4 MHz

#define SLEEP     255       // sleep for 255 cycles, so 2mn
#define INTERVAL  720       // 1440mn in 24h, sleep is 2mn long, so 720 intervals
#define WATER     45000     // watering duration 45 seconds

volatile uint8_t rtcIntSemaphore;  // flag from RTS interrupt that may be used by polPUMP function
uint16_t counter = 0;

void setup() {
  initSerialGPIO();     // initialize serial and GPIO

  init32kOscRTCPIT();   // init the 32K internal Osc and RTC-PIT for interrupts
  initSleepMode();      // set up the sleep mode
}

void loop() {
  while(1) {
    counter += 1;

    if( counter > INTERVAL ){
      counter = 0;

      ADC0_init();
      int humidity = ADC0_read();

      if( humidity < 50 ){
        pumpOn(); // turn on PUMP
        delay(WATER);
        pumpOff(); // turn off PUMP
      }
    }

    sleepNCycles(SLEEP); // cycles are about 500ms each / max 255 cycles so 2mn
  }
}

void ADC0_init(void) {
  // Disable digital input buffer
  PORTA.PIN2CTRL &= ~PORT_ISC_gm;
  PORTA.PIN2CTRL |= PORT_ISC_INPUT_DISABLE_gc;

  // Disable pull-up resistor 
  PORTA.PIN2CTRL &= ~PORT_PULLUPEN_bm; 

  ADC0.CTRLC =  ADC_PRESC_DIV4_gc       // CLK_PER divided by 4 
              | ADC_REFSEL_VDDREF_gc;   // VDD reference 
  ADC0.CTRLA =  ADC_RESSEL_10BIT_gc     // 10-bit mode 
              | ADC_ENABLE_bm;          // ADC Enable: enabled 
  ADC0.MUXPOS = ADC_MUXPOS_AIN2_gc;     // Select ADC channel 
}

uint16_t ADC0_read(void) {
  ADC0.COMMAND = ADC_STCONV_bm;                 // Start ADC conversion

  while ( !(ADC0.INTFLAGS & ADC_RESRDY_bm) );   // Wait until ADC conversion done 
  ADC0.INTFLAGS = ADC_RESRDY_bm;                // Clear the interrupt flag by writing 1

  return ADC0.RES;
}


//////////////////////////////////////////////////////////////////
//  ISR(RTC_PIT_vect)
//
//  Interrupt Service Routine for the RTC PIT interrupt
//
ISR(RTC_PIT_vect) {
  RTC.PITINTFLAGS = RTC_PI_bm;  // clear the interrupt flag   
  rtcIntSemaphore = 1;          // mark to PUMP function that interrupt has occurred

}


//////////////////////////////////////////////////////////////////
//  initSerialGPIO()
//
//  initialize all needed GPIO
//
void initSerialGPIO(void) {
  PORTA.DIRSET = PIN1_bm; //  set port to output for PUMP
  PORTA.OUTCLR = PIN1_bm; // turn off PUMP

  PORTA.DIRSET = PIN3_bm; //  set port to output for SENSOR
  PORTA.OUTSET = PIN3_bm; // turn on SENSOR
}


//////////////////////////////////////////////////////////////////
//  init32kOscRTCPIT()
//
//  initialize the internal ultra low power 32 kHz osc and periodic Interrupt timer
//
//  these two peripherals are interconnected in that the internal 32 kHz
//  osc will not start until a peripheral (PIT in this case) calls for it,
//  and the PIT interrupt should not be enabPUMP unit it is confirmed that
//  the 32 kHz osc is running and stable
//
//  Note that there is ERRATA on the RTC counter,  see:
//  https://ww1.microchip.com/downloads/en/DeviceDoc/ATtiny212-214-412-414-416-SilConErrataClarif-DS80000933A.pdf
//
//  That document currently states "Any write to the RTC.CTRLA register resets the 15-bit prescaler
//  resulting in a longer period on the current count or period".  So if you load the prescaler
//  value and then later enable the RTC (both on the same register) you get a very long
//  (max?) time period.  My solution to this problem is to always enable the RTC in the same
//  write that sets up the prescaler.  That seems to be an effective work-around. 
//
void init32kOscRTCPIT(void) {
  _PROTECTED_WRITE(CLKCTRL.OSC32KCTRLA, CLKCTRL_RUNSTDBY_bm);   // enable internal 32K osc   
  RTC.CLKSEL = RTC_CLKSEL_INT1K_gc;                             // Select 1.024 kHz from 32KHz Low Power Oscillator (OSCULP32K) as clock source
  RTC.PITCTRLA = RTC_PERIOD_CYC512_gc | RTC_PITEN_bm;           // Enable RTC-PIT with divisor of 512 (~500 milliseconds)
  while (!(CLKCTRL.MCLKSTATUS & CLKCTRL_OSC32KS_bm));           // wait until 32k osc clock has stabilized
  while (RTC.PITSTATUS > 0);                                    // wait for RTC.PITCTRLA synchronization to be achieved
  RTC.PITINTCTRL = RTC_PI_bm;                                   // allow interrupts from the PIT device  
}

//////////////////////////////////////////////////////////////////
//  initSleepMode()
//
//  this doesn't invoke sleep, it just sets up the type of sleep mode
//  to enter when the MCU is put to sleep by calling "sleep_cpu()"
//
void initSleepMode(void) {
  SLPCTRL.CTRLA = SLPCTRL_SMODE_PDOWN_gc;   // set sleep mode to "power down"
  SLPCTRL.CTRLA |= SLPCTRL_SEN_bm;          // enable sleep mode
}


//////////////////////////////////////////////////////////////////
//  sleepNCycles(uint16_t val)
//
//  cause system to go to sleep for N cycles, where each cycle
//  is about 500mS.  Note that processor does wakeup every 500 ms
//  but goes back to sleep almost immediately if it has not yet
//  done the desired number of cycles.
//
//  At 4 MHz MCU speed, the awake duration of each iternation in the
//  for loop below is approximately 1.5 microseconds.
//
void sleepNCycles(uint8_t val) {
  // first cycle may not be full cycle
  disableAllPeripherals(); // all off

  for (uint8_t i = 0; i < val ; i++) {
    sleep_cpu();  // put MCU to sleep
  }
  // now awake, sleep cycles complete, continue on
  initSerialGPIO(); // initialize serial and GPIO 
}


//////////////////////////////////////////////////////////////////
//  disableAllPeripherals()
//
//  To achieve minimum current consumption during sleep, it's best to
//  disable everything that might draw current during sleep.
//  This function disables everything except the RTC-PIT and 32K
//  internal low power oscillator.  This function does not disable
//  the counter used by the Arduino framework to implement millis(),
//  although the clock that drives that counter is suspended during sleep.
//
void disableAllPeripherals(void) {
  PORTA.DIRCLR = PIN0_bm; //  set port A0 to input
  PORTA.DIRCLR = PIN1_bm; //  set port A1 to input
  PORTA.DIRCLR = PIN2_bm; //  set port A2 to input
  PORTA.DIRCLR = PIN3_bm; //  set port A3 to input
  PORTA.DIRCLR = PIN6_bm; //  set port A4 to input
  PORTA.DIRCLR = PIN7_bm; //  set port A5 to input        
        
  PORTA.PIN0CTRL = PORT_ISC_INPUT_DISABLE_gc; // disable input buffers
  PORTA.PIN1CTRL = PORT_ISC_INPUT_DISABLE_gc;
  PORTA.PIN2CTRL = PORT_ISC_INPUT_DISABLE_gc;
  PORTA.PIN3CTRL = PORT_ISC_INPUT_DISABLE_gc;
  PORTA.PIN6CTRL = PORT_ISC_INPUT_DISABLE_gc;
  PORTA.PIN7CTRL = PORT_ISC_INPUT_DISABLE_gc;
}

void pumpOn(void)     {PORTA.OUTSET = PIN1_bm;} // turn on active high PUMP on PA1
void pumpOff(void)    {PORTA.OUTCLR = PIN1_bm;} // turn off active high PUMP on PA1
